/****************************************************************************
*
* (C) 2003 by BECK IPC GmbH
*
*  BECK IPC GmbH
*  Garbenheimerstr. 36-38
*  D-35578 Wetzlar
*
*  Phone : (49)-6441-905-0
*  Fax   : (49)-6441-905-245
*
* ---------------------------------------------------------------------------
* Module         : SPIAPI.C
* Function       : SPI API functions
* ---------------------------------------------------------------------------

$Header: SPIAPI.C, 12, 13.04.2004 11:34:09, Andre Pribil$

*****************************************************************************/

/***************************************************************************/
// Includes
/***************************************************************************/
#include <dos.h>
#include "spiapi.h"
#include "spi_priv.h"

/*
 @doc DOC_INDEX
 ---------------------------------------------------------------
 SPI API Table of Contents
 ---------------------------------------------------------------
 @clibidx  <tab><cfont>#include <InBraces SPIAPI\.H><efont> |

 @libcat1x  | SPI Init | spi90 |
    spi_init | Init the SPI Interface |
 @libcat2x  | SPI Access| spi95 |
    spi_write | Write to the SPI Interface |
 @libcat2x  | SPI Access | spi96 |
    spi_read | Read from the SPI Interface |
 @libcat2x  | SPI Access | spi97 |
    spi_read_write | Read and write from/to the SPI Interface |
*/

/***************************************************************************/
// Init SPI
/***************************************************************************/
#ifdef _spi90_
/*
 @doc DOC_spi90_
 ---------------------------------------------------------------
 Init SPI
 ---------------------------------------------------------------
 @clib  spi_init | Init the SPI Interface |
        <cfont>void spi_init ( unsigned int mode, char clock,<nl>
        <tab><tab><tab><tab>char dataIn, char dataOut ); <efont> |
 @parm  unsigned int  | mode  | must be '0'<nl>
 @parm  char  | clock | Number of the pio which should be used as clock pin
 @parm  char  | dataIn  | Number of the pio which should be used as data in pin
 @parm  char  | dataOut | Number of the pio which should be used as data out pin
 @rdesc -- none --.
 @Also  spi95.htm |  spi_write |
        spi96.htm |  spi_read  | |
 @rtos  This library function invokes a RTOS software interrupt.
        Refer to this RTOS API function's
        <linkto>../../i2c_spi.htm#0xAA0x90<linkmid>documentation<linktail>
        for more details.
*/
void spi_init( unsigned mode, char clock, char dataIn, char dataOut)
{
  _BX = mode;
  _CL = clock;
  _CH = dataIn;
  _DL = dataOut;
  _AH = SPI_INIT;
  asm int SPI_INT
}
#endif

/***************************************************************************/
// SPI Write
/***************************************************************************/
#ifdef _spi95_
/*
 @doc DOC_spi95_
 ---------------------------------------------------------------
 SPI Write
 ---------------------------------------------------------------
 @clib  spi_write | Write to the SPI Interface |
        <cfont>void spi_write ( const void far *bufferPtr, unsigned short nbytes );<efont> |
 @parm  void * | bufferPtr  | Pointer to buffer which contains write data
 @parm  int    | nbytes     | number of bytes to write
 @rdesc  -- none --.
 @Also  spi90.htm |  spi_init |
        spi96.htm |  spi_read | |
 @rtos  This library function invokes a RTOS software interrupt.
        Refer to this RTOS API function's
        <linkto>../../i2c_spi.htm#0xAA0x95<linkmid>documentation<linktail>
        for more details.
*/
void spi_write (const void far *bufferPtr, unsigned short nbytes)
{
  // write data
  _ES = FP_SEG(bufferPtr);
  _SI = FP_OFF(bufferPtr);
  _CX = nbytes;
  _AH = SPI_WRITE;
  asm int SPI_INT
}
#endif


/***************************************************************************/
// SPI Read
/***************************************************************************/
#ifdef _spi96_
/*
 @doc DOC_spi96_
 ---------------------------------------------------------------
 SPI Read
 ---------------------------------------------------------------
 @clib  spi_read | Read from the SPI Interface |
        <cfont>void spi_read ( void far *bufferPtr, unsigned short nbytes );<efont> |
 @parm  void * | bufferPtr  | Pointer to buffer where read data should be stored
 @parm  int    | nbytes     | number of bytes to read
 @rdesc  -- none --.
 @Also  spi90.htm |  spi_init  |
        spi95.htm |  spi_write | |
 @rtos  This library function invokes a RTOS software interrupt.
        Refer to this RTOS API function's
        <linkto>../../i2c_spi.htm#0xAA0x96<linkmid>documentation<linktail>
        for more details.
*/
void spi_read (void far *bufferPtr, unsigned short nbytes)
{
  // read data
  _ES = FP_SEG(bufferPtr);
  _DI = FP_OFF(bufferPtr);
  _CX = nbytes;
  _AH = SPI_READ;
  asm int SPI_INT
}
#endif



/***************************************************************************/
// SPI Read and Write
/***************************************************************************/
#ifdef _spi97_
/*
 @doc DOC_spi97_
 ---------------------------------------------------------------
 SPI Read
 ---------------------------------------------------------------
 @clib  spi_read_write | Read and write from the SPI Interface simultaneously |
        <cfont>void spi_read_write ( void far *rbufferPtr, const void far *wbufferPtr,
        <tab><tab><tab><tab><tab>&nbsp; unsigned short nbytes );<efont> |
 @parm  void * | rbufferPtr | Pointer to buffer where read data should be stored
 @parm  void * | wbufferPtr | Pointer to buffer which contains write data
 @parm  int    | nbytes     | number of bytes to read
 @rdesc  -- none --.
 @Also  spi90.htm |  spi_init  |
        spi95.htm |  spi_write |
        spi96.htm |  spi_read
 @rtos  This library function invokes a RTOS software interrupt.
        Refer to this RTOS API function's
        <linkto>../../i2c_spi.htm#0xAA0x97<linkmid>documentation<linktail>
        for more details.
*/
void _cdecl spi_read_write (void far *rbufferPtr, const void far *wbufferPtr, unsigned short nbytes)
{
  // read and write data
  asm push ds
  _DS = FP_SEG(rbufferPtr);
  _DI = FP_OFF(rbufferPtr);
  _ES = FP_SEG(wbufferPtr);
  _SI = FP_OFF(wbufferPtr);
  _CX = nbytes;
  _AH = SPI_READ_WRITE;
  asm int SPI_INT
  asm pop ds
}
#endif

