/******************************************************************************
*
* (c) 2003-2004 by BECK IPC GmbH
*
* BECK IPC GmbH
* Garbenheimerstr. 36-38
* D-35578 Wetzlar
*
* Phone: (49)-6441-905-0
* Fax:   (49)-6441-905-245
*
*******************************************************************************
*
* Disclaimer: This program is an example and should be used as such.
*             If you wish to use this program or parts of it in your application,
*             you must validate the code yourself. BECK IPC GmbH can not be held
*             responsible for the correct functioning or coding of this example.
* Module:     ftplib.c
* Function:   This module provides an API for the File Transfer Protocol.
* Course:     - Call ftpOpenControlConnection() and save socket descriptor
*             - Call ftpLogin()
*             - Call any manipulating functions you like (e.g. ftpReceiveFile())
*             - Call ftpLogout()
*
*******************************************************************************
*
* $Header: ftplib.c, 6, 20.04.2004 13:25:59, Jan Schatz$
*
******************************************************************************/



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <time.h>
#include <sys\stat.h>
#include <clib.h>
#include "bool.h"
#include "ftplib.h"



/******************************************************************************
* Constants
******************************************************************************/
//#define FTP_PRINT_MESSAGES  // Uncomment this to see FTP messages
//#define FTP_DEBUG           // Uncomment this to see function calls



/******************************************************************************
* ftpSendCommand()
*
* Sends a command string to the given (connected!) FTP server socket (control
* connection)
*
* Returns:
*  TRUE - Success
*  FALSE - Sending timed out
******************************************************************************/
static BOOL ftpSendCommand(
  int controlSocket,           // Socket to which the command is sent (control
                               // connection)
  const char *commandString )  // Command string that is to be sent
{
  unsigned int i,                      // Counter variable
               timeout = FTP_TIMEOUT;  // Timeout counter
  int result,
      error;                           // Result of send()-call
  #ifdef FTP_PRINT_MESSAGES
  char *tmpStringP;                    // Used to strip \r\n for output
  #endif


  // Check command string
  if( strlen( commandString ) == 0 )
    return FALSE;


  #ifdef FTP_DEBUG
  printf( "\r\n--> ftpSendCommand( %i, \"%c...\" );", controlSocket, commandString[0] );
  #endif


  // Send command
  for( i = 0; i < strlen( commandString ) && timeout > 0; )
  {
    if( ( result = send( controlSocket, commandString + i, min( strlen( commandString ) - i, INT_MAX ),
                         MSG_DONTWAIT, &error ) ) > 0 )
      i += result;
    else
    {
      RTX_Sleep_Time( 1000 );
      timeout -= ( timeout >= 1000 ) ? 1000 : timeout;
    }
  }

  // Check if sending timed out
  if( timeout == 0 )
    return FALSE;
  else
  {
    #ifdef FTP_PRINT_MESSAGES
    // Strip \r\n from command line
    while( ( tmpStringP = strchr( commandString, '\r' ) ) != NULL )
      *tmpStringP = '\0';
    while( ( tmpStringP = strchr( commandString, '\n' ) ) != NULL )
      *tmpStringP = '\0';

    printf( "\r\n>> %s", commandString );
    #endif
    return TRUE;
  }
}



/******************************************************************************
* ftpReceiveReply()
*
* Receives a reply from the given (connected!) FTP server socket (control
* connection)
*
* Returns:
*  1 - Success (this was the last line of the reply)
*  2 - Success (there are more lines of the reply)
*  FALSE - Receiving timed out
******************************************************************************/
static unsigned int ftpReceiveReply(
  int controlSocket,        // Socket from which the reply is received (control
                            // connection)
  char *replyString )       // Reply string - Must be large enough for
                            // FTP_SIZE_LINE charcters
{
  static unsigned int bufferSize = 0;   // Counter variable
  static char buffer[FTP_SIZE_LINE+1];  // Buffer for received data
  int result,
      error;                            // Results of recv()-call
  char *tmpStringP;                     // Temporary string pointer -
                                        // needed to check if the reply
                                        // line is complete, used to strip \r\n
                                        // for output
  unsigned int nRetries = 0;            // Number of retries


  #ifdef FTP_DEBUG
  printf( "\r\n--> ftpReceiveReply( %i, %p );", controlSocket, replyString );
  #endif


  // Receive the reply string
  while( TRUE )  // Condition see below
  {
    // Check if buffer contains a complete reply line
    if( ( tmpStringP = strstr( buffer, "\r\n" ) ) != NULL )
    {
      // Skip "\r\n"
      tmpStringP += 2;

      // Copy reply string to replyString
      strncpy( replyString, buffer, (size_t)( tmpStringP - buffer ) );
      replyString[(size_t)( tmpStringP - buffer )] = '\0';

      // Save rest of buffer
      bufferSize = strlen( buffer ) - (size_t)( tmpStringP - buffer );
      memmove( buffer, tmpStringP, bufferSize + 1 );

      // Leave
      break;
    }

    // This is the (negatived) loop condition.
    // It has to be placed here because the chunk of code above this has to be
    // executed at the beginning and the end of the loop.
    if( bufferSize >= FTP_SIZE_LINE )
      break;

    if( ( result = recv( controlSocket, buffer + bufferSize, min( FTP_SIZE_LINE - bufferSize, INT_MAX ),
                         MSG_TIMEOUT, FTP_TIMEOUT, &error ) ) > 0 )
    {
      bufferSize += result;
      nRetries = 0;

      // Make a string out of the buffer
      buffer[bufferSize] = '\0';
    }
    else
      if( ++nRetries == 3 )
        return FALSE;
  }

  // Strip \r\n from replyLine
  while( ( tmpStringP = strchr( replyString, '\r' ) ) != NULL )
    *tmpStringP = '\0';
  while( ( tmpStringP = strchr( replyString, '\n' ) ) != NULL )
    *tmpStringP = '\0';
    
  #ifdef FTP_PRINT_MESSAGES
  printf( "\r\n<< %s", replyString );
  #endif

  // Check if this was the last line of the reply
  if( strlen( replyString ) >= 6 && replyString[3] == ' ' )
    return 1;
  else
    return 2;
}



/******************************************************************************
* ftpSendCommandGetReplyCode()
*
* Sends a command string to the given (connected!) FTP server socket (control
* connection) and receives the reply from the FTP server
*
* Returns:
*  The Return code of the reply on success
*  FALSE - on error
******************************************************************************/
static unsigned int ftpSendCommandGetReplyCode(
  int controlSocket,          // Socket to which the command is sent (control
                              // connection)
  const char *commandString ) // Command string that is to be sent
{
  int result;                              // Results of recv()-call
  static char replyString[FTP_SIZE_LINE];  // Received reply string
  #ifdef DEBUG
  char *tmpStringP;                        // Used to strip \r\n for output
  #endif


  // Check command string
  if( strlen( commandString ) == 0 )
    return FALSE;


  #ifdef FTP_DEBUG
  printf( "\r\n--> ftpSendCommandGetReplyCode( %i, \"%c...\" );", controlSocket, commandString[0] );
  #endif


  // Send command
  if( !ftpSendCommand( controlSocket, commandString ) )
    return FALSE;

  // Receive reply
  while( ( result = ftpReceiveReply( controlSocket, replyString ) ) == 2 )
    ;
  if( !result )
    return FALSE;

  // Convert reply code and return it
  return atoi( replyString );
}



/******************************************************************************
* ftpSendFileToSocket()
*
* Sends the contents of a file to the given (connected!) FTP server socket
* (data connection)
* The socket will be closed after sending the file since stream mode is used.
*
* Returns:
*  TRUE - Success
*  FALSE - Sending timed out or file could not be read
******************************************************************************/
static BOOL ftpSendFileToSocket(
  int dataSocket,          // Socket to which the file is sent (data connection)
  const char *filename )   // Name of file on local system
{
  FILE *file;                           // File handle
  static char buffer[FTP_SIZE_BUFFER];  // Buffer to store data in
  unsigned int bufferSize,              // Number of bytes in buffer
               i,                       // Counter variable
               timeout;                 // Timeout counter
  int result,
      error;                            // Result of send()-call


  #ifdef FTP_DEBUG
  printf( "\r\n--> ftpSendFileToSocket( %i, \"%s\" );", dataSocket, filename );
  #endif


  // Open file
  if( ( file = fopen( filename, "rb" ) ) == NULL )
  {
    closesocket( dataSocket, &error );
    return FALSE;
  }

  // Get some data from file
  while( ( bufferSize = fread( (void *)buffer, 1, FTP_SIZE_BUFFER, file ) ) != 0 )
  {
    // Send data
    for( i = 0, timeout = FTP_TIMEOUT; i < bufferSize && timeout > 0; )
    {
      if( ( result = send( dataSocket, buffer + i, min( bufferSize - i, INT_MAX ),
                           MSG_DONTWAIT, &error ) ) > 0 )
        i += result;
      else
      {
        RTX_Sleep_Time( 1000 );
        timeout -= ( timeout >= 1000 ) ? 1000 : timeout;
      }
    }

    // Check for timeout
    if( timeout == 0 )
    {
      // Close file and socket
      fclose( file );
      closesocket( dataSocket, &error );

      return FALSE;
    }
  }

  // Close file and socket after transmission
  fclose( file );
  closesocket( dataSocket, &error );

  return TRUE;
}



/******************************************************************************
* ftpReceiveFileFromSocket()
*
* Receives data from the given (connected!) FTP server socket (data connection)
* and saves it to a file
* The socket will be closed after receiving the file since stream mode is used.
*
* Returns:
*  TRUE - Success
*  FALSE - Receiving timed out or file could not be written
******************************************************************************/
static BOOL ftpReceiveFileFromSocket(
  int dataSocket,          // Socket from which the file is received (data
                           // connection)
  const char *filename )   // Name of file on local system
{
  FILE *file;                           // File handle
  static char buffer[FTP_SIZE_BUFFER];  // Buffer to store data in
  int result,
      error;                            // Result of send()-call
  unsigned int nRetries = 0;            // Number of retries


  #ifdef FTP_DEBUG
  printf( "\r\n--> ftpReceiveFileFromSocket( %i, \"%s\" );", dataSocket, filename );
  #endif


  // Open file
  if( ( file = fopen( filename, "wb" ) ) == NULL )
  {
    closesocket( dataSocket, &error );
    return FALSE;
  }

  // Receive some data
  while( TRUE )
  {
    if( ( result = recv( dataSocket, buffer, min( FTP_SIZE_BUFFER, INT_MAX ),
                         MSG_TIMEOUT, FTP_TIMEOUT, &error ) ) > 0 )
    {
      nRetries = 0;

      // Write data to file
      if( fwrite( (void *)buffer, 1, result, file ) != result )
      {
        // Error
        fclose( file );
        closesocket( dataSocket, &error );
        return FALSE;
      }
    }
    else
    {
      if( ++nRetries == 3 )
        break;
    }
  }

  // Close file and socket after transmission
  fclose( file );
  closesocket( dataSocket, &error );

  return TRUE;
}



/******************************************************************************
* ftpOpenControlConnection()
*
* Establishs a control connection to a FTP server
*
* Returns:
*  The desciptor of the connected socket on success
*  -1 - Error
******************************************************************************/
int ftpOpenControlConnection(
  const char *remoteIp )  // IP address of remote station
{
  int controlSocket,                     // Socket descriptor of control connection
      result,                            // Result of ftpReceiveReply()
      error;                             // Result of API-calls
  static char tmpString[FTP_SIZE_LINE];  // Temporary string variable for replies
  struct sockaddr_in socketAddress;      // Address structure for socket


  #ifdef FTP_DEBUG
  printf( "\r\n--> ftpOpenControlConnection( \"%s\" );", remoteIp );
  #endif


  // Open socket
  if( ( controlSocket = opensocket( SOCK_STREAM, &error ) ) == -1 )
    return -1;

  // Prepare for connection
  socketAddress.sin_family = PF_INET;
  socketAddress.sin_port = htons( FTP_PORT );

  // Connect to server
  if( inet_addr( remoteIp, &socketAddress.sin_addr.s_addr ) != 0 ||
      connect( controlSocket, &socketAddress, &error ) != 0 )
  {
    closesocket( controlSocket, &error );
    return -1;
  }

  // Receive hello message
  while( ( result = ftpReceiveReply( controlSocket, tmpString ) ) == 2 )
    ;
  if( !result )
    return -1;

  // Analyse reply
  if( atoi( tmpString ) == 220 )
    return controlSocket;
  else
  return -1;
}



/******************************************************************************
* ftpOpenDataConnectionPassive()
*
* Opens a data connection to a FTP server
* Use this function if server operates in passive mode.
*
* Returns:
*  The desciptor of the connected socket on success
*  -1 - Error
******************************************************************************/
static int ftpOpenDataConnectionPassive(
  const char *remoteIp,      // IP address of remote station
  unsigned int dataPort )    // Port on remote station to connect to
{
  int dataSocket = 0,                // Socket descriptor
      error;                         // Result of API-calls
  struct sockaddr_in socketAddress;  // Address structure for socket


  #ifdef FTP_DEBUG
  printf( "\r\n--> ftpOpenDataConnectionPassive( \"%s\", %u );", remoteIp, dataPort );
  #endif

  
  // Open a socket
  if( ( dataSocket = opensocket( SOCK_STREAM, &error ) ) == -1 )
    return -1;

  // Prepare information for connection
  socketAddress.sin_family = PF_INET;
  socketAddress.sin_port = htons( dataPort );

  // Connect to server
  if( inet_addr( remoteIp, &socketAddress.sin_addr.s_addr ) != 0 ||
      connect( dataSocket, &socketAddress, &error ) != 0 )
  {
    closesocket( dataSocket, &error );
    return -1;
  }

  return dataSocket;
}



/******************************************************************************
* ftpOpenDataConnectionActive1()
*
* Opens a data connection to a FTP server (1. part)
* Use the functions ftpOpenDataConnectionActive1() and
* ftpOpenDataConnectionActive2() if server operates in active mode.
* - Call ftpOpenDataConnectionActive1() to create a socket that listens on a
*   random local port. The port number will be saved in port variable.
* - Make the server connect to this port by sending PORT and e.g. RETR commands.
* - Accept the server's data connection with a call to
*   ftpOpenDataConnectionActive2().
*
* Returns:
*  The desciptor of the listening socket on success
*  -1 - Error
******************************************************************************/
static int ftpOpenDataConnectionActive1(
  unsigned int *listenPort )  // Will be set to the port number of the listening
                              // socket
{
  int listenSocket = 0,              // Descriptor of listening socket
      error;                         // Result of API-calls
  struct sockaddr_in socketAddress;  // Address structure for socket
  time_t t;                          // Needed to generate random port number


  #ifdef FTP_DEBUG
  printf( "\r\n--> ftpOpenDataConnectionActive1( %p );", listenPort );
  #endif


  // Random socket
  RTX_Sleep_Time( 2000 );  // Wait more than one second to get different port
                           // numbers in subsequent calls
  srand( (unsigned int) time( &t ) );
  *listenPort = ( (unsigned int)rand() % (unsigned int)64512L ) + 1024;
  
  // Open a socket
  if( ( listenSocket = opensocket( SOCK_STREAM, &error ) ) == -1 )
    return -1;

  // Prepare information for connection
  socketAddress.sin_family = PF_INET;
  socketAddress.sin_addr.s_addr = 0L;
  socketAddress.sin_port = htons( *listenPort );

  // Bind socket and switch it to listening mode
  if( bind( listenSocket, &socketAddress, &error ) != 0 ||
      listen( listenSocket, 1, &error ) != 0 )
  {
    closesocket( listenSocket, &error );
    return -1;
  }

  // Return socket descriptor and local port
  return listenSocket;
}



/******************************************************************************
* ftpOpenDataConnectionActive2()
*
* Opens a data connection to a FTP server (2. part)
* Use the functions ftpOpenDataConnectionActive1() and
* ftpOpenDataConnectionActive2() if server operates in active mode.
* - Call ftpOpenDataConnectionActive1() to create a socket that listens on a
*   random local port. The port number will be saved in port variable.
* - Make the server connect to this port by sending PORT and e.g. RETR commands.
* - Accept the server's data connection with a call to
*   ftpOpenDataConnectionActive2().
*
* Returns:
*  The desciptor of the connected socket on success
*  -1 - Error
******************************************************************************/
static int ftpOpenDataConnectionActive2(
  int listenSocket )  // Socket that is listening for server connection
{
  int dataSocket = 0,                // Socket descriptor for data connection
      error;                         // Result of API-calls
  struct sockaddr_in socketAddress;  // Address structure for socket


  #ifdef FTP_DEBUG
  printf( "\r\n--> ftpOpenDataConnectionActive2( %i );", listenSocket );
  #endif


  // Accept server connection on socket
  socketAddress.sin_family = PF_INET;
  socketAddress.sin_addr.s_addr = 0L;
  socketAddress.sin_port = 0;

  if( ( dataSocket = accept( listenSocket, &socketAddress, &error ) ) == -1 )
    return -1; 

  // Close listening socket
  closesocket( listenSocket, &error );
  return dataSocket;  // Connection from FTP server
}



/******************************************************************************
* ftpLogin()
*
* Logs in on the given FTP server
* (USER, PASS, ACCT)
*
* Returns:
*  TRUE - on success
*  FALSE - on error
******************************************************************************/
BOOL ftpLogin(
  int controlSocket,     // Socket descriptor of control connection
  const char *user,
  const char *password,
  const char *account )  // Login information - each must not longer than
                         // FTP_SIZE_LINE - strlen( "XXXX \r\n" ) bytes
                         // If not needed with your server set to NULL
{
  enum { STATE_USER, STATE_PASS,
         STATE_ACCT, STATE_ERROR,
         STATE_OK
       } state = STATE_USER;             // Needed for the state machine that
                                         // implements the sequence of login commands
  static char tmpString[FTP_SIZE_LINE];  // Temporary string variable for commands
                                         // and replies


  #ifdef FTP_DEBUG
  printf( "\r\n--> ftpLogin( %i, \"%s\", \"%s\", \"%s\" );", controlSocket, user,
                                                             password, account );
  #endif


  // Check lengths of string parameters
  if( ( user && strlen( user ) > FTP_SIZE_LINE - strlen( "USER \r\n" ) ) ||
      ( password && strlen( password ) > FTP_SIZE_LINE - strlen( "PASS \r\n" ) ) ||
      ( account && strlen( account ) > FTP_SIZE_LINE - strlen( "ACCT \r\n" ) ) )
    return FALSE;

    
  // State machine
  while( state != STATE_ERROR &&
         state != STATE_OK )
  {
    switch( state )
    {
      case STATE_USER:  // Send USER command
        // Send command and analyse reply code
        sprintf( tmpString, "USER %s\r\n", user );
        switch( ftpSendCommandGetReplyCode( controlSocket, tmpString ) )
        {
          case 230:  // Logged in
            state = STATE_OK;
            break;

          case 331:  // Need password
            state = STATE_PASS;
            break;

          case 332:  // Need account
            state = STATE_ACCT;
            break;

          case 421: case 500: case 501: case 530:
          case 0:
          default:  // Error
            state = STATE_ERROR;
            break;
        }  // switch(...
        break;
         
      case STATE_PASS:  // Send PASS command
        // Send command and analyse reply code
        sprintf( tmpString, "PASS %s\r\n", password );
        switch( ftpSendCommandGetReplyCode( controlSocket, tmpString ) )
        {
          case 202: case 230:  // Logged in
            state = STATE_OK;
            break;

          case 332:  // Need account
            state = STATE_ACCT;
            break;

          case 421: case 500: case 501: case 503: case 530:
          case 0:
          default:  // Error
            state = STATE_ERROR;
            break;
        }  // switch(...
        break;

      case STATE_ACCT:  // Send ACCT command
        // Send command and analyse reply code
        sprintf( tmpString, "ACCT %s\r\n", account );
        switch( ftpSendCommandGetReplyCode( controlSocket, tmpString ) )
        {
          case 202: case 230:  // Logged in
            state = STATE_OK;
            break;

          case 421: case 500: case 501: case 503: case 530:
          case 0:
          default:  // Error
            state = STATE_ERROR;
            break;
        }  // switch(...
        break;
    }  // switch( state )
  }  // while( ...


  // Check if any error occured
  if( state == STATE_ERROR )
    return FALSE;
  else
    return TRUE;
}



/******************************************************************************
* ftpLogout()
*
* Logs out from the given FTP server and closes the socket
* (QUIT)
*
* Returns:
*  TRUE - on success
*  FALSE - on error
******************************************************************************/
BOOL ftpLogout(
  int controlSocket )  // Socket descriptor of control connection
{
  static char tmpString[FTP_SIZE_LINE];  // Temporary string variable for commands
                                         // and replies
  int error;                             // Result of API-calls


  #ifdef FTP_DEBUG
  printf( "\r\n--> ftpLogout( %i );", controlSocket );
  #endif


  // Send command and analyse reply code
  sprintf( tmpString, "QUIT\r\n" );
  if( ftpSendCommandGetReplyCode( controlSocket, tmpString ) == 221 )
  {
    if( closesocket( controlSocket, &error ) == -1 )
      return FALSE;
    else
      return TRUE;
  }
  else
    return FALSE;
}



/******************************************************************************
* ftpSetPort()
*
* Tells the FTP server which IP address and port number to connect to in active
* connection mode.
* (PORT)
*
* Returns:
*  TRUE - on success
*  FALSE - on error
******************************************************************************/
static BOOL ftpSetPort(
  int controlSocket,         // Socket descriptor of control connection
  unsigned int listenPort )  // Port to which the server is to connect
{
  static char tmpString[FTP_SIZE_LINE];  // Temporary string variable for command
  char localIp[16];                      // IP address of local station
  unsigned long lTmp;                    // Temporary variable

  
  #ifdef FTP_DEBUG
  printf( "\r\n--> ftpSetPort( %i, %u );", controlSocket, listenPort );
  #endif


  // Get local IP address
  Get_IPConfig( localIp, NULL, NULL );
  if( inet_addr( localIp, &lTmp ) != 0 )
    return FALSE;

  // Send PORT command and analyse reply code
  sprintf( tmpString, "PORT %u,%u,%u,%u,%u,%u\r\n",
           (unsigned int)( (unsigned char)lTmp ),
           (unsigned int)( (unsigned char)( lTmp >> 8 ) ),
           (unsigned int)( (unsigned char)( lTmp >> 16 ) ),
           (unsigned int)( (unsigned char)( lTmp >> 24 ) ),
           (unsigned int)( (unsigned char)( listenPort >> 8 ) ),
           (unsigned int)( (unsigned char)listenPort ) );
  if( ftpSendCommandGetReplyCode( controlSocket, tmpString ) != 200 )
    return FALSE;
  else
    return TRUE;
}



/******************************************************************************
* ftpSetPassive()
*
* Tells the FTP server to work in passive mode.
* (PASV)
*
* Returns:
*  TRUE - on success
*  FALSE - on error
******************************************************************************/
static BOOL ftpSetPassive(
  int controlSocket,          // Socket descriptor of control connection
  char *remoteIp,             // Will be set to the IP address the server will
                              // listen on
                              // Must be large enough for 16 bytes
  unsigned int *listenPort )  // Will be set to the port the server will listen on
{
  static char tmpString[FTP_SIZE_LINE];  // Temporary string variable for commands
                                         // and replies
  char *tmpStringP;                      // Temporary string pointer
  int result;                            // Result of ftpReceiveReply()-call
  unsigned int tmp[6];                   // Temporary variables


  #ifdef FTP_DEBUG
  printf( "\r\n--> ftpSetPassive( %i );", controlSocket );
  #endif


  // Send PASV command and analyse reply code
  sprintf( tmpString, "PASV\r\n" );
  if( !ftpSendCommand( controlSocket, tmpString ) )
    return FALSE;

  while( ( result = ftpReceiveReply( controlSocket, tmpString ) ) == 2 )
    ;
  if( !result )
    return -1;

  // Analyse reply
  if( strlen( tmpString ) < 3 ||
      atoi( tmpString ) != 227 )
    return FALSE;

  // Parse reply string
  tmpStringP = tmpString + 3;  // Skip reply code
  tmpStringP += strcspn( tmpStringP, "0123456789" );  // Skip any text before IP
  if( sscanf( tmpStringP, "%u,%u,%u,%u,%u,%u", &tmp[0], &tmp[1], &tmp[2], &tmp[3],
                                               &tmp[4], &tmp[5] ) != 6 ||
      tmp[0] > 255 ||
      tmp[1] > 255 ||
      tmp[2] > 255 ||
      tmp[3] > 255 ||
      tmp[4] > 255 ||
      tmp[5] > 255 )  // unsigned char
    return FALSE;

  // Return IP address and port
  sprintf( remoteIp, "%u.%u.%u.%u", tmp[0], tmp[1], tmp[2], tmp[3] );
  *listenPort = ( tmp[4] << 8 ) + tmp[5];
  return TRUE;
}



/******************************************************************************
* ftpSetType()
*
* Tells the FTP server which representation type is to be used for transfer.
* (TYPE)
*
* Returns:
*  TRUE - on success
*  FALSE - on error
******************************************************************************/
static BOOL ftpSetType(
  int controlSocket,  // Socket descriptor of control connection
  FTP_TYPE type )     // Representation type that is to be set
{
  static char tmpString[FTP_SIZE_LINE];  // Temporary string variable for commands
                                         // and replies


  #ifdef FTP_DEBUG
  printf( "\r\n--> ftpSetType( %i, %i );", controlSocket, type );
  #endif


  // Prepare TYPE command
  switch( type )
  {
    case FTP_TYPE_ASCII_NONPRINT:
      sprintf( tmpString, "TYPE A\r\n" );
      break;

    case FTP_TYPE_IMAGE:
      sprintf( tmpString, "TYPE I\r\n" );
      break;

    default:  // Illegal argument
      return FALSE;
  }

  // Send TYPE command and analyse reply code
  if( ftpSendCommandGetReplyCode( controlSocket, tmpString ) != 200 )
    return FALSE;
  else
    return TRUE;
}



/******************************************************************************
* ftpAllocate()
*
* Tells the server to allocate sufficient memory for a subsequent file transfer.
* (ALLO)
*
* Returns:
*  TRUE - on success (or if the server didn't understand the ALLO command and thus
*                     doesn't seem to need it)
*  FALSE - on error
******************************************************************************/
static BOOL ftpAllocate(
  int controlSocket,    // Socket descriptor of control connection
  unsigned long size )  // Number of (logical) bytes that is to be allocated
{
  static char tmpString[FTP_SIZE_LINE];  // Temporary string variable for command


  #ifdef FTP_DEBUG
  printf( "\r\n--> ftpAllocate( %i, %lu );", controlSocket, size );
  #endif


  // Send ALLO command and analyse reply code
  sprintf( tmpString, "ALLO %lu\r\n", size );
  switch( ftpSendCommandGetReplyCode( controlSocket, tmpString ) )
  {
    case 200: case 202:  // OK
      return TRUE;
    case 500: // Server doesn't understand ALLO command. Assume it isn't needed.
      return TRUE;
    case 421: case 501: case 504: case 530:
    case 0:
    default:  // Error
      return FALSE;
  }
}



/******************************************************************************
* ftpListDirectory()
*
* Lists the content of the given diredctory on the remote system
* (LIST)
*
* Returns:
*  TRUE - on success
*  FALSE - on error
******************************************************************************/
BOOL ftpListDirectory(
  int controlSocket,           // Socket descriptor of control connection
  const char *directory,       // Directory whose content is to be listed
                               // Must not be longer than
                               // FTP_SIZE_LINE - strlen( "LIST \r\n" ) bytes
  FILE *destination,           // Normaly this should be set to stdout to print
                               // the listing to the console.
  BOOL passive )               // If TRUE passive mode will be used
{
  int listenSocket,                      // Socket descriptor for listening socket
                                         // (needed with active mode)
      dataSocket,                        // Socket descriptor for data connection
      result,
      error;                             // Result of API-calls
  static char tmpString[FTP_SIZE_LINE];  // Temporary string variable for commands
                                         // and replies
  unsigned int listenPort;               // Port number of listening socket
                                         // (local or remote)
  char remoteIp[16];                     // IP address of server - for passive mode


  #ifdef FTP_DEBUG
  printf( "\r\n--> ftpListDirectory( %i, %p, %u );", controlSocket, destination,
                                                     (unsigned int)passive );
  #endif


  // Check length of directory
  if( strlen( directory ) > FTP_SIZE_LINE - strlen( "LIST /\r\n" ) )
    return FALSE;                                      // ^ might be added


  // Set representation type
  if( !ftpSetType( controlSocket, FTP_TYPE_ASCII_NONPRINT ) )
    return FALSE;


  // Prepare data connection
  if( passive )  // Passive mode
  {
    // Set passive mode
    if( !ftpSetPassive( controlSocket, remoteIp, &listenPort ) )
      return FALSE;
  }
  else  // Active mode
  {
    // Open listening socket
    if( ( listenSocket = ftpOpenDataConnectionActive1( &listenPort ) ) == -1 )
      return FALSE;

    // Set port
    if( !ftpSetPort( controlSocket, listenPort ) )
    {
      closesocket( listenSocket, &error );
      return FALSE;
    }
  }


  // Send LIST command and analyse reply code
  if( strlen( directory ) > 0 &&
      directory[strlen( directory )-1] != '/' )
    // Add '/' if directory != "" and there isn't already one
    sprintf( tmpString, "LIST %s/\r\n", directory );
  else
    sprintf( tmpString, "LIST %s\r\n", directory );   
  if( ftpSendCommandGetReplyCode( controlSocket, tmpString ) != 150 )
  {
    if( !passive )
      closesocket( listenSocket, &error );
    return FALSE;
  }


  // Establish data connection
  if( passive )  // Passive mode
  {
    // Open connection
    if( ( dataSocket = ftpOpenDataConnectionPassive( remoteIp, listenPort ) ) == -1 )
      return FALSE;
  }
  else  // Active mode
  {
    // Accept connection
    if( ( dataSocket = ftpOpenDataConnectionActive2( listenSocket ) ) == -1 )
    {
      closesocket( listenSocket, &error );
      return FALSE;
    }
  }


  // Receive directory listing
  while( ( result = ftpReceiveReply( dataSocket, tmpString ) ) != FALSE )
    fprintf( destination, "\r\n%s", tmpString );



  // Receive and analyse second reply
  while( ( result = ftpReceiveReply( controlSocket, tmpString ) ) == 2 )
    ;
  if( !result )
    return FALSE;                                

  switch( atoi( tmpString ) )
  {
    case 226: case 250:  // OK
      return TRUE;

    case 425: case 426: case 451:
    case 0:
    default:  // Error
      return FALSE;                           
  }
}



/******************************************************************************
* ftpCreateDirectory()
*
* Creates a directory on the remote system
* (MKD)
*
* Returns:
*  TRUE - on success
*  FALSE - on error
******************************************************************************/
BOOL ftpCreateDirectory(
  int controlSocket,       // Socket descriptor of control connection
  const char *directory )  // Name of new directory
                           // Must not be longer than
                           // FTP_SIZE_LINE - strlen( "MKD \r\n" ) bytes
{
  static char tmpString[FTP_SIZE_LINE];  // Temporary string variable for command


  #ifdef FTP_DEBUG
  printf( "\r\n--> ftpCreateDirectory( %i, \"%s\" );", controlSocket, directory );
  #endif


  // Check directory
  if( directory == NULL ||
      strlen( directory ) > FTP_SIZE_LINE - strlen( "MKD \r\n" ) )
    return FALSE;


  // Send MKD command and analyse reply code
  sprintf( tmpString, "MKD %s\r\n", directory );
  if( ftpSendCommandGetReplyCode( controlSocket, tmpString ) != 257 )
    return FALSE;

  return TRUE;
}



/******************************************************************************
* ftpDeleteDirectory()
*
* Deletes a directory on the remote system
* (RMD)
*
* Returns:
*  TRUE - on success
*  FALSE - on error
******************************************************************************/
BOOL ftpDeleteDirectory(
  int controlSocket,       // Socket descriptor of control connection
  const char *directory )  // Directory which is to be deleted
                           // Must not be longer than
                           // FTP_SIZE_LINE - strlen( "RMD \r\n" ) bytes
{
  static char tmpString[FTP_SIZE_LINE];  // Temporary string variable for command


  #ifdef FTP_DEBUG
  printf( "\r\n--> ftpDeleteDirectory( %i, \"%s\" );", controlSocket, directory );
  #endif


  // Check directory
  if( directory == NULL ||
      strlen( directory ) > FTP_SIZE_LINE - strlen( "RMD \r\n" ) )
    return FALSE;


  // Send RMD command and analyse reply code
  sprintf( tmpString, "RMD %s\r\n", directory );
  if( ftpSendCommandGetReplyCode( controlSocket, tmpString ) != 250 )
    return FALSE;

  return TRUE;
}



/******************************************************************************
* ftpSetDirectory()
*
* Changes the working directory on the remote system.
* (CWD, CDUP)
*
* Returns:
*  TRUE - on success
*  FALSE - on error
******************************************************************************/
BOOL ftpSetDirectory(
  int controlSocket,       // Socket descriptor of control connection
  const char *directory )  // Directory which is to be set as working directory
                           // Must not be longer than
                           // FTP_SIZE_LINE - strlen( "CWD \r\n" ) bytes
                           // Can be ".." to reach parent directory
{
  static char tmpString[FTP_SIZE_LINE];  // Temporary string variable for command


  #ifdef FTP_DEBUG
  printf( "\r\n--> ftpSetDirectory( %i, \"%s\" );", controlSocket, directory );
  #endif


  // Check directory
  if( directory == NULL ||
      strlen( directory ) > FTP_SIZE_LINE - strlen( "CWD \r\n" ) )
    return FALSE;


  // Send CWD or CDUP command and analyse reply code
  if( strcmp( directory, ".." ) == 0 )
    sprintf( tmpString, "CDUP\r\n" );
  else
  {
    if( strlen( directory ) > 0 &&
        directory[strlen( directory )-1] != '/' )
      // Add '/' if directory != "" and there isn't already one
      sprintf( tmpString, "CWD %s/\r\n", directory );
    else
      sprintf( tmpString, "CWD %s\r\n", directory );
  }
  switch( ftpSendCommandGetReplyCode( controlSocket, tmpString ) )
  {
    case 200: case 250:  // OK
      return TRUE;

    case 421: case 500: case 501: case 502: case 530: case 550:
    case 0:
    default:  // Error
      return FALSE;
  }
}



/******************************************************************************
* ftpGetDirectory()
*
* Gets the working directory on the remote system.
* (PWD)
*
* Returns:
*  TRUE - on success
*  FALSE - on error
******************************************************************************/
BOOL ftpGetDirectory(
  int controlSocket,       // Socket descriptor of control connection
  char *directory )        // Will be set to the current working directory
                           // Must be large enough for FTP_SIZE_LINE bytes
{
  static char tmpString[FTP_SIZE_LINE];  // Temporary string variable for command
                                         // and reply
  char *tmpStringP;                      // Needed to parse reply string
  int result;                            // Result of ftpReceiveReply()-calls


  #ifdef FTP_DEBUG
  printf( "\r\n--> ftpGetDirectory( %i, \"%p\" );", controlSocket, directory );
  #endif


  // Send PWD command and analyse reply
  sprintf( tmpString, "PWD\r\n" );
  if( !ftpSendCommand( controlSocket, tmpString ) )
    return FALSE;

  while( ( result = ftpReceiveReply( controlSocket, tmpString ) ) == 2 )
    ;
  if( !result )
    return -1;

  // Analyse reply
  if( strlen( tmpString ) < 5 ||
      atoi( tmpString ) != 257 )
    return FALSE;

  // Parse reply string
  if( ( tmpStringP = strchr( tmpString, '"' ) ) == NULL )  // Find start of directory
    return FALSE;
  ++tmpStringP;  // Skip '"'
  strcpy( directory, tmpStringP );  // Copy directory and junk after it
  if( ( tmpStringP = strrchr( directory, '"' ) ) == NULL )  // Find end of directory
    return FALSE;
  *tmpStringP = '\0';  // Discard junk after directory

  return TRUE;
}



/******************************************************************************
* ftpSendFile()
*
* Sends a given file to the FTP server and tells it ti store it on the remote
* system
* (STOR)
*
* Returns:
*  TRUE - on success (This does not indicate whether the file was sent
*                     completely. One cannot determine that in stream mode.)
*  FALSE - on error
******************************************************************************/
BOOL ftpSendFile(
  int controlSocket,           // Socket descriptor of control connection
  const char *filenameRemote,  // Name of file on remote system
                               // Must not be longer than
                               // FTP_SIZE_LINE - strlen( "STOR \r\n" ) bytes
  const char *filenameLocal,   // Name of file on local system
  FTP_TYPE type,               // Representation type that is to be used
  BOOL passive )               // If TRUE passive mode will be used
{
  int listenSocket,                      // Socket descriptor for listening socket
                                         // (needed with active mode)
      dataSocket,                        // Socket descriptor for data connection
      result,
      error;                             // Result of API-calls
  static char tmpString[FTP_SIZE_LINE];  // Temporary string variable for commands
                                         // and replies
  unsigned int listenPort;               // Port number of listening socket
                                         // (local or remote)
  char remoteIp[16];                     // IP address of server - for passive mode
  struct stat fileInfo;                  // Information on the local file


  #ifdef FTP_DEBUG
  printf( "\r\n--> ftpSendFile( %i, \"%s\", \"%s\", %u );", controlSocket,
                                                            filenameRemote,
                                                            filenameLocal,
                                                            (unsigned int)passive );
  #endif


  // Check length of remote filename
  if( filenameRemote == NULL ||
      strlen( filenameRemote ) > FTP_SIZE_LINE - strlen( "STOR \r\n" ) )
    return FALSE;


  // Set representation type
  if( !ftpSetType( controlSocket, type ) )
    return FALSE;


  // Get information on local file and allocate sufficient memory on remote system
  if( stat( filenameLocal, &fileInfo ) == -1 ||
      !ftpAllocate( controlSocket, fileInfo.st_size ) )
    return FALSE;

       
  // Prepare data connection
  if( passive )  // Passive mode
  {
    // Set passive mode
    if( !ftpSetPassive( controlSocket, remoteIp, &listenPort ) )
      return FALSE;
  }
  else  // Active mode
  {
    // Open listening socket
    if( ( listenSocket = ftpOpenDataConnectionActive1( &listenPort ) ) == -1 )
      return FALSE;

    // Set port
    if( !ftpSetPort( controlSocket, listenPort ) )
    {
      closesocket( listenSocket, &error );
      return FALSE;
    }
  }


  // Send STOR command and analyse reply code
  sprintf( tmpString, "STOR %s\r\n", filenameRemote );
  if( ftpSendCommandGetReplyCode( controlSocket, tmpString ) != 150 )
  {
    if( !passive )
      closesocket( listenSocket, &error );
    return FALSE;
  }


  // Establish data connection
  if( passive )  // Passive mode
  {
    // Open connection
    if( ( dataSocket = ftpOpenDataConnectionPassive( remoteIp, listenPort ) ) == -1 )
      return FALSE;
  }
  else  // Active mode
  {
    // Accept connection
    if( ( dataSocket = ftpOpenDataConnectionActive2( listenSocket ) ) == -1 )
    {
      closesocket( listenSocket, &error );
      return FALSE;
    }
  }


  // Send file data
  if( !ftpSendFileToSocket( dataSocket, filenameLocal ) )
    return FALSE;


  // Receive and analyse second reply
  while( ( result = ftpReceiveReply( controlSocket, tmpString ) ) == 2 )
    ;
  if( !result )
    return FALSE;                                

  switch( atoi( tmpString ) )
  {
    case 226: case 250:  // OK
      return TRUE;

    case 110: case 425: case 426: case 451: case 551: case 552:
    case 0:
    default:  // Error
      return FALSE;                           
  }   
}



/******************************************************************************
* ftpReceiveFile()
*
* Retrieves a given file from the FTP server and stores it on the local file
* system
* (RETR)
*
* Returns:
*  TRUE - on success (This does not indicate whether the file was received
*                     completely. One cannot determine that in stream mode.)
*  FALSE - on error
******************************************************************************/
BOOL ftpReceiveFile(
  int controlSocket,           // Socket descriptor of control connection
  const char *filenameLocal,   // Name of file on local system
  const char *filenameRemote,  // Name of file on remote system
                               // Must not be longer than
                               // FTP_SIZE_LINE - strlen( "RETR \r\n" ) bytes
  FTP_TYPE type,               // Representation type that is to be used
  BOOL passive )               // If TRUE passive mode will be used
{
  int listenSocket,                      // Socket descriptor for listening socket
                                         // (needed with active mode)
      dataSocket,                        // Socket descriptor for data connection
      result,
      error;                             // Result of API-calls
  static char tmpString[FTP_SIZE_LINE];  // Temporary string variable for commands
                                         // and replies
  unsigned int listenPort;               // Port number of listening socket
                                         // (local or remote)
  char remoteIp[16];                     // IP address of server - for passive mode


  #ifdef FTP_DEBUG
  printf( "\r\n--> ftpReceiveFile( %i, \"%s\", \"%s\", %u );", controlSocket,
                                                               filenameLocal,
                                                               filenameRemote,
                                                               (unsigned int)passive );
  #endif


  // Check length of remote filename
  if( filenameRemote == NULL ||
      strlen( filenameRemote ) > FTP_SIZE_LINE - strlen( "RETR \r\n" ) )
    return FALSE;


  // Set representation type
  if( !ftpSetType( controlSocket, type ) )
    return FALSE;


  // Prepare data connection
  if( passive )  // Passive mode
  {
    // Set passive mode
    if( !ftpSetPassive( controlSocket, remoteIp, &listenPort ) )
      return FALSE;
  }
  else  // Active mode
  {
    // Open listening socket
    if( ( listenSocket = ftpOpenDataConnectionActive1( &listenPort ) ) == -1 )
      return FALSE;

    // Set port
    if( !ftpSetPort( controlSocket, listenPort ) )
    {
      closesocket( listenSocket, &error );
      return FALSE;
    }
  }


  // Send RETR command and analyse reply code
  sprintf( tmpString, "RETR %s\r\n", filenameRemote );
  if( ftpSendCommandGetReplyCode( controlSocket, tmpString ) != 150 )
  {
    if( !passive )
      closesocket( listenSocket, &error );
    return FALSE;
  }


  // Establish data connection
  if( passive )  // Passive mode
  {
    // Open connection
    if( ( dataSocket = ftpOpenDataConnectionPassive( remoteIp, listenPort ) ) == -1 )
      return FALSE;
  }
  else  // Active mode
  {
    // Accept connection
    if( ( dataSocket = ftpOpenDataConnectionActive2( listenSocket ) ) == -1 )
    {
      closesocket( listenSocket, &error );
      return FALSE;
    }
  }


  // Receive file data
  if( !ftpReceiveFileFromSocket( dataSocket, filenameLocal ) )
    return FALSE;


  // Receive and analyse second reply
  while( ( result = ftpReceiveReply( controlSocket, tmpString ) ) == 2 )
    ;
  if( !result )
    return FALSE;                                

  switch( atoi( tmpString ) )
  {
    case 226: case 250:  // OK
      return TRUE;

    case 110: case 425: case 426: case 451:
    case 0:
    default:  // Error
      return FALSE;                           
  }   
}



/******************************************************************************
* ftpDelete()
*
* Deletes a file on the remote system
* (DELE)
*
* Returns:
*  TRUE - on success
*  FALSE - on error
******************************************************************************/
BOOL ftpDelete(
  int controlSocket,      // Socket descriptor of control connection
  const char *filename )  // Filename
                          // Must not be longer than
                          // FTP_SIZE_LINE - strlen( "DELE \r\n" ) bytes
{
  static char tmpString[FTP_SIZE_LINE];  // Temporary string variable for command


  #ifdef FTP_DEBUG
  printf( "\r\n--> ftpDelete( %i, \"%s\" );", controlSocket, filename );
  #endif


  // Check filename
  if( filename == NULL ||
      strlen( filename ) > FTP_SIZE_LINE - strlen( "DELE \r\n" ) )
    return FALSE;


  // Send DELE command and analyse reply code
  sprintf( tmpString, "DELE %s\r\n", filename );
  if( ftpSendCommandGetReplyCode( controlSocket, tmpString ) != 250 )
    return FALSE;

  return TRUE;
}



/******************************************************************************
* ftpRename()
*
* Renames a file or directory on the remote system
* (RNFR, RNTO)
*
* Returns:
*  TRUE - on success
*  FALSE - on error
******************************************************************************/
BOOL ftpRename(
  int controlSocket,        // Socket descriptor of control connection
  const char *filenameNew,  // New name of file or directory
                            // Must not be longer than
                            // FTP_SIZE_LINE - strlen( "RNFR \r\n" ) bytes
  const char *filenameOld ) // Current name of file or directory
                            // Must not be longer than
                            // FTP_SIZE_LINE - strlen( "RNTO \r\n" ) bytes
{
  static char tmpString[FTP_SIZE_LINE];  // Temporary string variable for commands


  #ifdef FTP_DEBUG
  printf( "\r\n--> ftpRename( %i, \"%s\", \"%s\" );", controlSocket, filenameNew, filenameOld );
  #endif


  // Check names
  if( filenameOld == NULL ||
      strlen( filenameOld ) > FTP_SIZE_LINE - strlen( "RNFR \r\n" ) ||
      filenameNew == NULL ||
      strlen( filenameNew ) > FTP_SIZE_LINE - strlen( "RNTO \r\n" ) )
    return FALSE;


  // Send RNFR command and analyse reply code
  sprintf( tmpString, "RNFR %s\r\n", filenameOld );
  if( ftpSendCommandGetReplyCode( controlSocket, tmpString ) != 350 )
    return FALSE;

  // Send RNTO command and analyse reply code
  sprintf( tmpString, "RNTO %s\r\n", filenameNew );
  if( ftpSendCommandGetReplyCode( controlSocket, tmpString ) != 250 )
    return FALSE;

  return TRUE;
}



// End of file

