/******************************************************************************
*
* (c) 2003-2004 by BECK IPC GmbH
*
* BECK IPC GmbH
* Garbenheimerstr. 36-38
* D-35578 Wetzlar
*
* Phone: (49)-6441-905-0
* Fax:   (49)-6441-905-245
*
*******************************************************************************
*
* Disclaimer: This program is an example and should be used as such.
*             If you wish to use this program or parts of it in your application,
*             you must validate the code yourself. BECK IPC GmbH can not be held
*             responsible for the correct functioning or coding of this example.
*
* Module:     pop3.c
* Function:   This program is a simple POP3 client that connects to a given POP3
*             server, shows the number of messages in the given maildrop, fetches
*             the first message and deletes it if the user whishes.
*             The IP address of the POP3 server and the user and password for the
*             maildrop are expected as command line parameters. To get a usage
*             message call the program without any parameters.
* Notes:      To get a better grasp of what is done in the function fetchMail()
*             you should gallop through RFC 1939.
*
*******************************************************************************
*
* $Header: pop3.c, 2, 31.03.2004 11:15:52, Andre Pribil$
*
******************************************************************************/

/******************************************************************************
* Includes
******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>
#include <limits.h>
#include <clib.h>
#include "bool.h"



/******************************************************************************
* Constants
******************************************************************************/
//#define PRINT_MESSAGES    // Comment this out if you don't want to see any server
                          // messages.
#define POP3_PORT   110   // Well known POP3 port
#define BUFFERSIZE  2048  // Since there aren't many definitions for buffersizes
#define TIMEOUT     60    // and timeouts in RFC 1939 we take values that seem
                          // to be appropriate.



/******************************************************************************
* sendCommand()
*
* Sends a command string to the given (connected!) POP3 server socket.
*
* Returns:
*  1 - Success
*  0 - Sending timed out
******************************************************************************/
int sendCommand( int socket,                 // Socket to which the command is sent
                 const char *commandString,  // Command string
                 unsigned int *timeout )     // Timeout in seconds - Passed by
                                             // pointer so the caller can use
                                             // one timeout value for two commands
{
  unsigned int i;  // Counter variable
  int result,
      error;       // Result of send()-call


  // Send command
  for( i = 0; i < strlen( commandString ) && *timeout > 0; --*timeout )
  {
    if( ( result = send( socket, commandString + i, min( strlen( commandString ) - i, INT_MAX ),
                         MSG_DONTWAIT, &error ) ) > 0 )
      i += result;

    RTX_Sleep_Time( 1000 );
  }

  // Check if sending timed out
  if( *timeout == 0 )
    return 0;
  else
  {
    #ifdef PRINT_MESSAGES
    printf( "\r\n>> %s", commandString );
    #endif

    return 1;
  }
}



/******************************************************************************
* receiveReply()
*
* Receives a reply from the given (connected!) POP3 server socket.
*
* Returns:
*  1 - Success (there may be more lines, if this is a multi-line reply)
*  2 - Success (there are no more lines in this multi-line reply)
*  0 - Receiving timed out
******************************************************************************/
int receiveReply( int socket,               // Socket to which the command is sent
                  char *replyString,        // Command string - Must be large
                                            // enough for MAXSIZE_COMMANDLINE
                                            // charcters
                  unsigned int *timeout )   // Timeout in seconds - Passed by
                                            // pointer so the caller can use
                                            // one timeout value for two commands
{
  static unsigned int bufferSize = 0;      // Counter variable
  static char buffer[BUFFERSIZE+1];

  int result,
      error;           // Results of recv()-call
  char *tmpString;     // Temporary string pointer - needed to check if the reply
                       // string is complete

  // Receive the reply string
  while( 1 )  // Condition see below
  {
    // Check if buffer contains a complete reply string
    if( ( tmpString = strstr( buffer, "\r\n" ) ) != NULL )
    {
      // Skip "\r\n"
      tmpString += 2;

      // Copy reply string to replyString
      strncpy( replyString, buffer, (size_t)( tmpString - buffer ) );
      replyString[(size_t)( tmpString - buffer )] = '\0';

      // Save rest of buffer
      bufferSize = strlen( buffer ) - (size_t)( tmpString - buffer );
      memmove( buffer, tmpString, bufferSize + 1 );

      // Leave
      break;
    }

    // This is the (negatived) loop condition.
    // It has to be placed here because the chunk of code above this has to be
    // executed at the beginning and the end of the loop.
    if( bufferSize >= BUFFERSIZE || *timeout < 1 )
      break;

    if( ( result = recv( socket, buffer + bufferSize, min( BUFFERSIZE - bufferSize, INT_MAX ),
                         MSG_DONTWAIT, 0, &error ) ) > 0 )
    {
      bufferSize += result;

      // Make a string out of the buffer
      buffer[bufferSize] = '\0';
    }

    RTX_Sleep_Time( 1000 );
    --*timeout;
  }

  // Check if receiving timed out
  if( *timeout == 0 )
  {
    return 0;
  }
  else
  {
    #ifdef PRINT_MESSAGES
    printf( "\r\n<< %s", replyString );
    #endif

    // Check if this was the last line of a multi-line reply
    if( strlen( replyString ) >= 3 && strncmp( replyString, ".\r\n", 3 ) == 0 )
      return 2;
    else
    {
      if( replyString[0] == '.' )
        memmove( replyString, replyString + 1, strlen( replyString ) );
      return 1;
    }
  }
}



/******************************************************************************
* fetchMail()
*
* Prints a list of all messages in the given maildrop and prints the first
* message to STOUT.
*
* Returns:
*  TRUE - Success
*  FALSE - Error occured
******************************************************************************/
BOOL fetchMail(
               char *remoteIp,             // IP address of SMTP server
               char *user,
               char *password )            // Login data for the maildrop
{
  enum {
         STATE_CONNECTED, STATE_USER, STATE_PASS, STATE_STAT, STATE_RETR,
         STATE_DELE, STATE_RSET, STATE_QUIT, STATE_DISCONNECTED
       } state = STATE_DISCONNECTED;        // Needed for the state machine that
                                            // implements the sequence of POP3
                                            // commands
  BOOL errorOccured = FALSE;                // Will be set to TRUE if any error
                                            // occures
  char commandString[BUFFERSIZE+1],         // Buffers for commands and their replies
       replyString[BUFFERSIZE+1];           // + 1 byte for trailing '\0'
  struct sockaddr_in socketAddress;         // Address structure for SMTP server
                                            // socket
  unsigned int timeout,                     // Timeout for sendCommand() and
                                            // receiveReply()
               nMessages;                   // Number of messages in the maildrop
  int socket,                               // Socket descriptor
      result,                               // Result of receiveReply()
      error;                                // Error number returned by TCP/IP-API



  // Gather information for connection
  socketAddress.sin_family = AF_INET;
  socketAddress.sin_port = htons( POP3_PORT );
  if( inet_addr( remoteIp, &socketAddress.sin_addr.s_addr ) != 0 )
    return FALSE;

  // Open socket and connect
  printf( "\r\nConnecting, please wait..." );
  
  if( ( socket = opensocket( SOCK_STREAM, &error ) ) == -1 ||
        connect( socket, &socketAddress, &error ) != 0 )
  {
    errorOccured = TRUE;
    
    printf( "\r\nERROR: Got no POP3 connection." );
  }
  else
    state = STATE_CONNECTED;


  // State machine
  // Normal flow:
  // STATE_CONNECTED -> STATE_USER -> STATE_PASS ->
  // STATE_RETR -> STATE_DELE -> STATE_QUIT -> STATE_DISCONNECTED
  while( state != STATE_DISCONNECTED )
  {
    switch( state )
    {
      // Get first message from server
      case STATE_CONNECTED:
        timeout = TIMEOUT;
        if( !receiveReply( socket, replyString, &timeout ) ||
            replyString[0] != '+' )
        {
          errorOccured = TRUE;
          state = STATE_DISCONNECTED;

          printf( "\r\nERROR: Got no POP3 connection." );
        }
        else
          state = STATE_USER;
        
        break;


      // Send USER command
      case STATE_USER:
        sprintf( commandString, "USER %s\r\n", user );
        timeout = TIMEOUT;
        if( !sendCommand( socket, commandString, &timeout ) ||
            !receiveReply( socket, replyString, &timeout ) ||
            replyString[0] == '-' )
        {
          errorOccured = TRUE;
          state = STATE_DISCONNECTED;

          printf( "\r\nERROR: Could not log in." );
        }
        else
          state = STATE_PASS;

        break;


      // Send PASS command
      case STATE_PASS:
        sprintf( commandString, "PASS %s\r\n", password );
        timeout = TIMEOUT;
        if( !sendCommand( socket, commandString, &timeout ) ||
            !receiveReply( socket, replyString, &timeout ) ||
            replyString[0] == '-' )
        {
          errorOccured = TRUE;
          state = STATE_DISCONNECTED;

          printf( "\r\nERROR: Could not log in." );
        }
        else
          state = STATE_STAT;

        break;


      // Send STAT command
      case STATE_STAT:
        sprintf( commandString, "STAT\r\n" );
        timeout = TIMEOUT;
        if( !sendCommand( socket, commandString, &timeout ) ||
            !( result = receiveReply( socket, replyString, &timeout ) ) ||
            replyString[0] == '-' ||
            sscanf( replyString, "+OK %u", &nMessages ) == 0 )
        {
          errorOccured = TRUE;
          state = STATE_DISCONNECTED;

          printf( "\r\nERROR: Could not state number of messages in maildrop." );
        }
        else
        {
          printf( "\r\nThe maildrop contains %u messag%s.", nMessages,
                                                            ( nMessages == 1 ) ? "e" : "es" );

          if( nMessages > 0 )
            state = STATE_RETR;
          else
            state = STATE_QUIT;
        }

        break;


      // Send RETR command
      case STATE_RETR:
        sprintf( commandString, "RETR 1\r\n" );
        timeout = TIMEOUT;
        if( !sendCommand( socket, commandString, &timeout ) ||
            !( result = receiveReply( socket, replyString, &timeout ) ) ||
            replyString[0] == '-' )
        {
          errorOccured = TRUE;
          state = STATE_DISCONNECTED;

          printf( "\r\nERROR: Could not fetch message #1.", result );
        }
        else
        {
          printf( "\r\nFetching message #1."
                  "\r\nPress any key to continue..." );
          getch();

          printf( "\r\n-------------- Message #1 --------------"
                  "\r\n" );

          for( timeout = TIMEOUT;
               ( result = receiveReply( socket, replyString, &timeout ) ) == 1;
               timeout = TIMEOUT )
            printf( "%s", replyString );

          if( !result )
          {
            errorOccured = TRUE;
            state = STATE_DISCONNECTED;

            printf( "\r\nERROR: Could not fetch message #1." );
          }
          else
            state = STATE_DELE;

          printf( "\r\n----------------------------------------" );
        }

        break;


      // Send DELE command
      case STATE_DELE:
        printf( "\r\nPress 1 to delete message #1 now."
                "\r\nPress any other key to continue without deleting the message..." );
        if( getch() != '1' )
        {
          state = STATE_QUIT;
          break;
        }

        printf( "\r\nDeleting message #1..." );
        
        sprintf( commandString, "DELE 1\r\n" );
        timeout = TIMEOUT;
        if( !sendCommand( socket, commandString, &timeout ) ||
            !receiveReply( socket, replyString, &timeout ) ||
            replyString[0] == '-' )
        {
          errorOccured = TRUE;
          state = STATE_DISCONNECTED;

          printf( "\r\nERROR: Could not delete message #1" );
        }
        else
          state = STATE_QUIT;

        break;


      // Send QUIT command
      case STATE_QUIT:
      default:
        sprintf( commandString, "QUIT\r\n" );
        timeout = TIMEOUT;
        if( !sendCommand( socket, commandString, &timeout ) ||
            !receiveReply( socket, replyString, &timeout ) ||
            replyString[0] == '-' )
          errorOccured = TRUE;

        state = STATE_DISCONNECTED;

        break;
    }
  }


  // Close socket
  closesocket( socket, &error );

  return !errorOccured;
}



/******************************************************************************
* main()
******************************************************************************/
#pragma argsused
int main( int argc, char *argv[] )
{
  // Set STDIO focus to application
  BIOS_Set_Focus( FOCUS_APPLICATION );

  
  // Say hello
  printf( "\r\n--------------------------------------------------"
          "\r\nIPC@Chip POP3-Demo"
          "\r\n--------------------------------------------------"
          "\r\n" );


  // Check number of command line parameters
  if( argc < 4 )
  {
    printf( "\r\nUsage: SMTP.EXE REMOTEIP USER PASSWORD"
            "\r\n REMOTEIP:  IP address of POP3 server"
            "\r\n USER &"
            "\r\n PASSWORD:  Login data fro maildrop"
            "\r\n" );
  }
  else
    fetchMail( argv[1], argv[2], argv[3] );

  // Say bye
  printf( "\r\nGood Bye."
          "\r\n" );

  // Set STDIO focus to both
  BIOS_Set_Focus( FOCUS_BOTH );


  return 0;
}



// End of file