/******************************************************************************
*
* (c) 2004 by BECK IPC GmbH
*
* BECK IPC GmbH
* Garbenheimerstr. 36-38
* D-35578 Wetzlar
*
* Phone: (49)-6441-905-0
* Fax:   (49)-6441-905-245
*
*******************************************************************************
*
* Disclaimer: This program is an example and should be used as such.
*             If you wish to use this program or parts of it in your application,
*             you must validate the code yourself. BECK IPC GmbH can not be held
*             responsible for the correct functioning or coding of this example.
*
* Module:     smtp.c
* Function:   This program is a simple SMTP client that connects to a given SMTP
*             server and sends an E-Mail. The IP address of the SMTP server,
*             the sender address and at least one recipient address are expected
*             as command line parameters. To get a usage message call the program
*             without any parameters.
* Notes:      To get a better grasp of what is done in the function sendMail()
*             you should gallop through RFC 2821 (especially chapter 4).
*
*******************************************************************************
*
* $Header: smtp.c, 5, 31.03.2004 11:15:59, Andre Pribil$
*
******************************************************************************/


/******************************************************************************
* Includes
******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <mem.h>
#include <limits.h>
#include <clib.h>
#include "bool.h"
#include "smtp.h"



/******************************************************************************
* Constants
******************************************************************************/
#define PRINT_MESSAGES  // Comment this out if you don't want to see any server
                        // messages.



/******************************************************************************
* sendCommand()
*
* Sends a command string to the given (connected!) SMTP server socket.
*
* Returns:
*  1 - Success
*  0 - Command string is too long or sending timed out
******************************************************************************/
int sendCommand( int socket,                 // Socket to which the command is sent
                 const char *commandString,  // Command string
                 unsigned int *timeout )     // Timeout in seconds - Passed by
                                             // pointer so the caller can use
                                             // one timeout value for two commands
{
  unsigned int i;  // Counter variable
  int result,
      error;       // Result of send()-call


  // Send command
  for( i = 0; i < strlen( commandString ) && *timeout > 0; --*timeout )
  {
    if( ( result = send( socket, commandString + i, min( strlen( commandString ) - i, INT_MAX ),
                       MSG_DONTWAIT, &error ) ) > 0 )
      i += result;

    RTX_Sleep_Time( 1000 );
  }

  // Check if sending timed out
  if( *timeout == 0 )
    return 0;
  else
  {
    #ifdef PRINT_MESSAGES
    printf( "%s", commandString );
    #endif
    return 1;
  }
}


/******************************************************************************
* receiveReply()
*
* Receives a reply from the given (connected!) SMTP server socket.
*
* Returns:
*  1 - Success (this was the last line of the reply)
*  2 - Success (there are more lines of the reply)
*  0 - Receiving timed out
******************************************************************************/
int receiveReply( int socket,               // Socket from which the reply is received
                  char *replyString,        // Reply string - Must be large
                                            // enough for MAXSIZE_REPLYLINE
                                            // charcters
                  unsigned int *timeout )   // Timeout in seconds - Passed by
                                            // pointer so the caller can use
                                            // one timeout value for two commands
{
  static unsigned int bufferSize = 0;      // Counter variable
  static char buffer[MAXSIZE_REPLYLINE+1];

  int result,
      error;           // Results of recv()-call
  char *tmpString;     // Temporary string pointer - needed to check if the reply
                       // line is complete


  // Receive the reply string
  while( TRUE )  // Condition see below
  {
    // Check if buffer contains a complete reply line
    if( ( tmpString = strstr( buffer, "\r\n" ) ) != NULL )
    {
      // Skip "\r\n"
      tmpString += 2;

      // Copy reply string to replyString
      strncpy( replyString, buffer, (size_t)( tmpString - buffer ) );
      replyString[(size_t)( tmpString - buffer )] = '\0';

      // Save rest of buffer
      bufferSize = strlen( buffer ) - (size_t)( tmpString - buffer );
      memmove( buffer, tmpString, bufferSize + 1 );

      // Leave
      break;
    }

    // This is the (negatived) loop condition.
    // It has to be placed here because the chunk of code above this has to be
    // executed at the beginning and the end of the loop.
    if( bufferSize >= MAXSIZE_REPLYLINE || *timeout < 1 )
      break;

    if( ( result = recv( socket, buffer + bufferSize, min( MAXSIZE_REPLYLINE - bufferSize, INT_MAX ),
                         MSG_DONTWAIT, 0, &error ) ) > 0 )
    {
      bufferSize += result;

      // Make a string out of the buffer
      buffer[bufferSize] = '\0';
    }

    RTX_Sleep_Time( 1000 );
    --*timeout;
  }
    
  // Chek if receiving timed out
  if( *timeout == 0 )
    return 0;
  else
  {
    #ifdef PRINT_MESSAGES
    printf( "%s", replyString );
    #endif
      
    // Check if this was the last line of the reply
    if( strlen( replyString ) >= 6 && replyString[3] == ' ' )
      return 1;
    else
      return 2;
  }
}



/******************************************************************************
* sendMail()
*
* Sends an E-Mail to several senders.
*
* Returns:
*  TRUE - Success
*  FALSE - Error occured
******************************************************************************/
BOOL sendMail(
               char *remoteIp,             // IP address of SMTP server
               char *fromAddress,          // E-Mail address of sender
               unsigned int nToAddresses,  // Number of addresses in toAddresses[]
               char *toAddresses[],        // E-Mail address(es) of recipients
               char *mailData )            // Message that is going to be sent
                                           // Must be MIME-encoded and end
                                           // with "\r\n". One line must not be
                                           // longer than 1000 characters.
{
  enum { STATE_CONNECTED, STATE_EHLO, STATE_MAIL,
         STATE_RCPT, STATE_DATA_0, STATE_DATA_1,
         STATE_RSET, STATE_QUIT, STATE_DISCONNECTED
       } state = STATE_DISCONNECTED;        // Needed for the state machine that
                                            // implements the sequence of SMTP
                                            // commands
  BOOL errorOccured = FALSE;                // Will be set to TRUE if any error
                                            // occures
  char localIp[16],                         // IP address of this IPC@Chip
       commandLine[MAXSIZE_COMMANDLINE+1],  // Buffers for commands and their replies
       replyString[MAXSIZE_REPLYLINE+1];    // + 1 byte for trailing '\0'
  struct sockaddr_in socketAddress;         // Address structure for SMTP server
                                            // socket
  unsigned int timeout,                     // Timeout for sendCommand() and
                                            // receiveReply()
               nTries = 0,                  // Number of tries - max. 5
               iToAddresses = 0;            // Counter variable for toAddresses[]
  int socket,                               // Socket descriptor
      result,                               // Result of receiveReply()
      error;                                // Error number returned by TCP/IP-API


  // Gather information for connection
  Get_IPConfig( localIp, NULL, NULL );
  socketAddress.sin_family = AF_INET;
  socketAddress.sin_port = htons( SMTP_PORT );
  if( inet_addr( remoteIp, &socketAddress.sin_addr.s_addr ) != 0 )
    return FALSE;

  // Open socket and connect
  if( ( socket = opensocket( SOCK_STREAM, &error ) ) == -1 ||
        connect( socket, &socketAddress, &error ) != 0 )
    errorOccured = TRUE;
  else
    state = STATE_CONNECTED;


  // State machine
  // Normal flow:
  // STATE_CONNECTED -> STATE_EHLO -> STATE_MAIL -> STATE_RCPT ->
  // STATE_DATA_0 -> STATE_DATA_1 -> STATE_QUIT -> STATE_DISCONNECTED
  while( state != STATE_DISCONNECTED )
  {
    switch( state )
    {
      // Get first message from server
      case STATE_CONNECTED:
        timeout = TIMEOUT_CONNECT;
        if( !receiveReply( socket, replyString, &timeout ) )
        {
          errorOccured = TRUE;
          state = STATE_RSET;
          break;
        }

        switch( replyString[0] )
        {
          case '2':  // OK
            state = STATE_EHLO;
            break;
          case '5':  // Permanent error - Abort - Abort
          default:
            errorOccured = TRUE;
            state = STATE_QUIT;
            break;
        }
        break;


      // Send EHLO command
      case STATE_EHLO:
        sprintf( commandLine, "EHLO [%s]\r\n", localIp );
        timeout = TIMEOUT;
        if( !sendCommand( socket, commandLine, &timeout ) )
        {
          errorOccured = TRUE;
          state = STATE_RSET;
          break;
        }

        while( ( result = receiveReply( socket, replyString, &timeout ) ) == 2 );
        if( !result )
        {
          errorOccured = TRUE;
          state = STATE_RSET;
          break;
        }

        switch( replyString[0] )
        {
          case '2':  // OK - go to next state
            state = STATE_MAIL;
            break;
          case '5':  // Permanent error - Abort
          default:
            errorOccured = TRUE;
            state = STATE_QUIT;
            break;
        }
        break;


      // Send MAIL command
      case STATE_MAIL:
        if( nTries++ == 5 )
        {
          state = STATE_QUIT;
          break;
        }

        sprintf( commandLine, "MAIL FROM: <%s>\r\n", fromAddress );
        timeout = TIMEOUT_MAIL;
        if( !sendCommand( socket, commandLine, &timeout ) )
        {
          errorOccured = TRUE;
          state = STATE_RSET;
          break;
        }

        while( ( result = receiveReply( socket, replyString, &timeout ) ) == 2 );
        if( !result )
        {
          errorOccured = TRUE;
          state = STATE_RSET;
          break;
        }     

        switch( replyString[0] )
        {
          case '2':  // OK - go to next state
            state = STATE_RCPT;
            break;
          case '4':  // (Temporary) error - retry
            state = STATE_RSET;
            break;
          case '5':  // Permanent error - Abort
          default:
            errorOccured = TRUE;
            state = STATE_QUIT;
            break;
        }
        break;


      // Send RCPT command
      // This state will be processes once for every recipient
      case STATE_RCPT:
        sprintf( commandLine, "RCPT TO: <%s>\r\n", toAddresses[iToAddresses] );
        timeout = TIMEOUT_RCPT;
        if( !sendCommand( socket, commandLine, &timeout ) )
        {
          errorOccured = TRUE;
          state = STATE_RSET;
          break;
        }

        while( ( result = receiveReply( socket, replyString, &timeout ) ) == 2 );
        if( !result )
        {
          errorOccured = TRUE;
          state = STATE_RSET;
          break;
        }

        switch( replyString[0] )
        {
          case '2':  // OK - If this was the last to-address goto next state,
                     // else only increment iToAddresses to process next to-address.
            if( ++iToAddresses >= nToAddresses )
              state = STATE_DATA_0;
            break;
          case '4':  // (Temporary) error - retry
            state = STATE_RSET;
            break;
          case '5':  // Permanent error - Abort
          default:
            errorOccured = TRUE;
            state = STATE_QUIT;
            break;
        }
        break;


      // Send DATA command
      case STATE_DATA_0:
        sprintf( commandLine, "DATA\r\n" );
        timeout = TIMEOUT_DATAINITIATION;
        if( !sendCommand( socket, commandLine, &timeout ) )
        {
          errorOccured = TRUE;
          state = STATE_RSET;
          break;
        }

        while( ( result = receiveReply( socket, replyString, &timeout ) ) == 2 );
        if( !result )
        {
          errorOccured = TRUE;
          state = STATE_RSET;
          break;
        }

        switch( replyString[0] )
        {
          case '3': // Waiting for mail data
            state = STATE_DATA_1;
            break;
          case '4':  // (Temporary) error - retry
            state = STATE_RSET;
            break;
          case '5':  // Permanent error - Abort
          default:
            errorOccured = TRUE;
            state = STATE_QUIT;
            break;
        }
        break;


      // Send mail data
      case STATE_DATA_1:
        sprintf( commandLine, ".\r\n" );
        // Assuming that a chunk of data is equal to the size of the TCP send queue
        timeout = TIMEOUT_DATABLOCK * ( ( strlen( (char *)mailData ) / 4096 ) + 1 );
        if( !sendCommand( socket, (char *)mailData, &timeout ) ||
            !sendCommand( socket, commandLine, &timeout ) )
        {
          errorOccured = TRUE;
          state = STATE_RSET;
          break;
        }

        timeout = TIMEOUT_DATATERMINATION;
        while( ( result = receiveReply( socket, replyString, &timeout ) ) == 2 );
        if( !result )
        {
          errorOccured = TRUE;
          state = STATE_RSET;
          break;
        }

        switch( replyString[0] )
        {
          case '2':  // OK - go to next state
            state = STATE_QUIT;
            break;
          case '4':  // (Temporary) error - retry
            state = STATE_RSET;
            break;
          case '5':  // Permanent error - Abort
          default:
            errorOccured = TRUE;
            state = STATE_QUIT;
            break;
        }
        break;


      //Send RSET command
      case STATE_RSET:
        sprintf( commandLine, "RSET\r\n" );
        timeout = TIMEOUT;
        if( !sendCommand( socket, commandLine, &timeout ) )
        {
          errorOccured = TRUE;
          state = STATE_QUIT;
          break;
        }

        while( ( result = receiveReply( socket, replyString, &timeout ) ) == 2 );
        if( !result )
        {
          errorOccured = TRUE;
          state = STATE_RSET;
          break;
        }

        switch( replyString[0] )
        {
          case '2':  // OK - go to next state
            state = STATE_MAIL;
            break;
          default:
            errorOccured = TRUE;
            state = STATE_QUIT;
            break;
        }
        break;


      // Send QUIT command
      case STATE_QUIT:
      default:
        sprintf( commandLine, "QUIT\r\n" );
        timeout = TIMEOUT;
        if( !sendCommand( socket, commandLine, &timeout ) )
          errorOccured = TRUE;

        while( ( result = receiveReply( socket, replyString, &timeout ) ) == 2 );
        if( !result )
          errorOccured = TRUE;

        state = STATE_DISCONNECTED;
        break;
    }
  }


  // Close socket
  closesocket( socket, &error );

  return !errorOccured;
}



/******************************************************************************
* main()
******************************************************************************/
#pragma argsused
int main( int argc, char *argv[] )
{
  static char mailData[MAXSIZE_MESSAGECONTENT+1];  // Buffer for the whole message
                                                   // text + 1 byte for trailing
                                                   // '\0'


  // Set STDIO focus to application
  BIOS_Set_Focus( FOCUS_APPLICATION );

  // Say hello
  printf( "\r\n--------------------------------------------------"
          "\r\nIPC@Chip SMTP-Demo"
          "\r\n--------------------------------------------------"
          "\r\n" );


  // Check number of command line parameters
  if( argc < 4 )
  {
    printf( "\r\nUsage: SMTP.EXE REMOTEIP FROMADDRESS TOADDRESS [...]"
            "\r\n REMOTEIP:     IP address of SMTP server"
            "\r\n FROMADDRESS:  E-Mail address of the sender"
            "\r\n TOADDRESS:    E-Mail address(es) of the recipient"
            "\r\n               The frist recipient will be added to the mail header."
            "\r\n               All other recipients will get a blind copy."
            "\r\n" );
  }
  else
  {
    // Create mail body
    // Must be MIME-encoded and end with "\r\n".
    // One line must not be longer than 1000 characters.
    sprintf( (char *)mailData, "From: %s\r\n"
                               "Subject: IPC@Chip SMTP-Demo\r\n"
                               "To: %s\r\n"
                               "Dear recipient,\r\n"
                               "\r\n"
                               "This E-Mail was sent by the IPC@Chip SMTP-Demo.\r\n"
                               "\r\n"
                               "\r\n"
                               "Kind regards\r\n"
                               "Beck IPC GmbH\r\n"
                               "\r\n"
                               "------------------------------\r\n"
                               "http://www.beck-ipc.com\r\n",
                               argv[2], argv[3] );

    // Send
    printf( "\r\nSending message...", argv[3] );

    #ifdef PRINT_MESSAGES
    printf( "\r\n----------- Server messages ------------"
            "\r\n" );
    #endif

    // Note that a command line is limited to 80 characters!
    sendMail( argv[1], argv[2], argc - 3, &argv[3], mailData );

    #ifdef PRINT_MESSAGES
    printf( "\r\n----------------------------------------" );
    #endif
  }


  // Say bye
  printf( "\r\nGoodbye."
          "\r\n" );

  // Set STDIO focus to both
  BIOS_Set_Focus( FOCUS_BOTH );


  return 0;
}



// End of file
